import unittest
from unittest.mock import patch

from common.backend.models.base import LLMStep
from common.solutions.chains.image_generation.image_generation_chain import ImageGenerationChain


class TestImageGenerationChain(unittest.TestCase):

    @patch('common.solutions.chains.image_generation.image_generation_chain.get_num_images_user_can_generate')
    @patch('common.solutions.chains.image_generation.image_generation_chain.get_nbr_images_to_generate')
    @patch('common.solutions.chains.image_generation.image_generation_chain.NoRetrievalChain')
    @patch('common.solutions.chains.image_generation.image_generation_chain.dataiku_api')
    @patch('common.solutions.chains.image_generation.image_generation_chain.DKULLM')
    def test_run_image_generation_query_limit_reached(
        self,
        mock_llm,
        mock_dataiku_api,
        mock_NoRetrievalChain,
        mock_get_nbr_images_to_generate,
        mock_get_num_images_user_can_generate
    ):
        # Mock configurations and behaviors
        mock_dataiku_api.webapp_config = {"max_images_per_user_per_week": 5}
        mock_get_num_images_user_can_generate.return_value = 0
        mock_get_nbr_images_to_generate.return_value = 3
        mock_no_retrieval_instance = mock_NoRetrievalChain.return_value
        mock_no_retrieval_instance.get_as_json.return_value = {"answer": "max_images_per_user_per_week_reached"}

        # Prepare test inputs
        user_query = "Generate an image"
        referred_image = None
        user_profile = {"media": {"image": {"nbr_images_to_generate": {"value": 3}}}}

        # Initialize the ImageGenerationChain
        llm =  mock_llm.return_value
        image_gen_chain = ImageGenerationChain(llm, user_query, referred_image, user_profile)

        # Execute the method
        results = list(image_gen_chain.run_image_generation_query())

        # Validate results
        self.assertEqual(results[0], {"step": LLMStep.STREAMING_END})
        self.assertEqual(results[1], {"answer": "max_images_per_user_per_week_reached"})
        mock_get_num_images_user_can_generate.assert_called_with(user_profile)
        mock_NoRetrievalChain.assert_called_with(llm)
        mock_no_retrieval_instance.get_as_json.assert_called_once()

    @patch('common.solutions.chains.image_generation.image_generation_chain.get_num_images_user_can_generate')
    @patch('common.solutions.chains.image_generation.image_generation_chain.set_nbr_images_to_generate')
    @patch('common.solutions.chains.image_generation.image_generation_chain.dataiku_api')
    @patch('common.solutions.chains.image_generation.image_generation_chain.DKULLM')
    def test_run_image_generation_requested_images_more_than_quota_left(
        self,
        mock_llm,
        mock_dataiku_api,
        mock_set_nbr_images_to_generate,
        mock_get_num_images_user_can_generate
    ):
        # Mock configurations and behaviors
        mock_dataiku_api.webapp_config = {"max_images_per_user_per_week": 5}
        mock_get_num_images_user_can_generate.return_value = 3

        # Prepare test inputs
        user_query = "Generate an image"
        referred_image = None
        user_profile = {"media": {"image": {"nbr_images_to_generate": {"value": 4}}}}

        # Initialize the ImageGenerationChain
        llm =  mock_llm.return_value
        image_gen_chain = ImageGenerationChain(llm, user_query, referred_image, user_profile)

        # Execute the method
        results = list(image_gen_chain.run_image_generation_query())

        # Validate results
        new_user_profile = {"media": {"image": {"nbr_images_to_generate": {"value": 3}}}}
        mock_set_nbr_images_to_generate.return_value = new_user_profile
        mock_get_num_images_user_can_generate.assert_called_with(user_profile)
        mock_set_nbr_images_to_generate.assert_called_with(user_profile, 3)
        self.assertEqual(results[0], {"step": LLMStep.GENERATING_IMAGE})

